#include "BasicLightParam.h"
#include "CUDAPreintegratedLut.h"

static __device__ float d_GetScalarValue(const float4& pos);
static __device__ float4 d_GetNormal(const float4& pos);

texture<float4,2> TFTex1;
texture<float4,2> TFTex2;

#ifndef __BasicTFParam__
#define __BasicTFParam__
__constant__ float c_tf_scale;
__constant__ float c_tf_shift;
#endif 


#include <cuda_runtime.h> 
#include <cutil.h>
#include <cutil_math.h> 

__constant__ float c_sampleDistance;
__constant__ int c_shade;
__constant__ float3 c_ambient;
__constant__ float3 c_diffuse;
__constant__ float3 c_specular;
__constant__ float c_specular_power;
__constant__ float4 c_light_direction;
__constant__ float c_spec_threshold;

__constant__ float3 c_BackgroundColor;

static __device__ 
void d_VolumeModel_Shade(const float4& pos,const float4& CameraDirection,float3 &k,float3 &b)
{
	float4 normal=d_GetNormal(pos);
	float n_dot_c;
	n_dot_c=dot(normal,CameraDirection);
	
	float n_dot_l;
	float n_dot_h; 
	
	k=c_ambient;
	b=make_float3(0.0f); 
	n_dot_l=dot(normal,c_light_direction);
	
	if ((n_dot_l>0.0f && n_dot_c>0.0f) || (n_dot_l<0.0f && n_dot_c<0.0f))
	{
		n_dot_l=n_dot_l>0?n_dot_l:-n_dot_l;
		k+=c_diffuse*n_dot_l;
		float4 HalfwayVector=normalize(c_light_direction+CameraDirection);
		n_dot_h=dot(normal,HalfwayVector);
		n_dot_h=n_dot_h>0?n_dot_h:-n_dot_h;
		if (n_dot_h > c_spec_threshold)
		{
			b=c_specular*powf(n_dot_h,c_specular_power);
		}
	}
}


static bool FullRayCasterPrepare( bool b_shade, float SampleDistance, const CUDAPreintegratedLut* lut, const BasicLightParam* param)
{
	if (!param) return false;

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_sampleDistance, &SampleDistance, sizeof(float),0,cudaMemcpyHostToDevice));
	
	int shade=b_shade?1:0;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_shade, &shade, sizeof(int),0,cudaMemcpyHostToDevice));
	
	float lightColor[3];
	lightColor[0]=param->m_LightColor[0];
	lightColor[1]=param->m_LightColor[1];
	lightColor[2]=param->m_LightColor[2];

	float LightIntensity=param->m_LightIntensity;

	lightColor[0]*=LightIntensity;
	lightColor[1]*=LightIntensity;
	lightColor[2]*=LightIntensity;

	float f_ambient=param->m_Ambient;
	float ambient[3];
	ambient[0]=f_ambient*lightColor[0];
	ambient[1]=f_ambient*lightColor[1];
	ambient[2]=f_ambient*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_ambient, ambient, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float f_diffuse=param->m_Diffuse;
	float diffuse[3];
	diffuse[0]=f_diffuse*lightColor[0];
	diffuse[1]=f_diffuse*lightColor[1];
	diffuse[2]=f_diffuse*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_diffuse, diffuse, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float f_specular=param->m_Specular;
	float specular[3];
	specular[0]=f_specular*lightColor[0];
	specular[1]=f_specular*lightColor[1];
	specular[2]=f_specular*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_specular, specular, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float specular_power=param->m_SpecularPower;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_specular_power, &specular_power, sizeof(float),0,cudaMemcpyHostToDevice));

	float spec_threshold=powf(0.001f/max(max(specular[0],specular[1]),specular[2]),1.0f/specular_power);
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_spec_threshold, &spec_threshold, sizeof(float),0,cudaMemcpyHostToDevice));

	float light_direction[4];
	light_direction[0]=param->m_LightDirection[0];
	light_direction[1]=param->m_LightDirection[1];
	light_direction[2]=param->m_LightDirection[2];
	light_direction[3]=0.0f;

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_light_direction, light_direction, sizeof(float)*4,0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_BackgroundColor, param->m_BackgroundColor, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	int sampleNum=lut->m_SampleNumber;
	float minPos=lut->m_min;
	float maxPos=lut->m_max;

	float scale=(sampleNum-1)/(maxPos-minPos);
	float shift=-(sampleNum-1)*minPos/(maxPos-minPos)+0.5f;
	
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_tf_scale, &scale, sizeof(float),0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_tf_shift, &shift, sizeof(float),0,cudaMemcpyHostToDevice));

	cudaChannelFormatDesc channelDesc;

	TFTex1.normalized = false; 
	TFTex1.filterMode = cudaFilterModePoint;
    TFTex1.addressMode[0] = cudaAddressModeClamp;  
	TFTex1.addressMode[1] = cudaAddressModeClamp;  

	CUDA_SAFE_CALL(cudaGetChannelDesc(&channelDesc, lut->m_d_TFArray1));
	CUDA_SAFE_CALL(cudaBindTextureToArray(TFTex1,lut->m_d_TFArray1, channelDesc));

	TFTex2.normalized = false; 
	TFTex2.filterMode = cudaFilterModePoint;
    TFTex2.addressMode[0] = cudaAddressModeClamp;  
	TFTex2.addressMode[1] = cudaAddressModeClamp;  

	CUDA_SAFE_CALL(cudaGetChannelDesc(&channelDesc, lut->m_d_TFArray2));
	CUDA_SAFE_CALL(cudaBindTextureToArray(TFTex2, lut->m_d_TFArray2, channelDesc));

	return true;
}

static void FullRayCasterClear()
{
	CUDA_SAFE_CALL(cudaUnbindTexture(TFTex1));
	CUDA_SAFE_CALL(cudaUnbindTexture(TFTex2));
}



