#include "BasicLightParam.h"
#include "IsosurfaceRayCasterParam.h"

static __device__ float d_GetScalarValue(const float4& pos);
static __device__ float4 d_GetNormal(const float4& pos);

#include <cuda_runtime.h> 
#include <cutil.h>
#include <cutil_math.h>

__constant__ float c_isovalue;
__constant__ float4 c_baseColor;
__constant__ float3 c_ambient;
__constant__ float3 c_diffuse;
__constant__ float3 c_specular;
__constant__ float c_specular_power;
__constant__ float4 c_light_direction;
__constant__ float c_spec_threshold;

__constant__ int c_Illustrative;
__constant__ float3 c_BackgroundColor;

static __device__ 
float4 d_GetColor(const float4& pos,const float4& CameraDirection,bool increasing)
{
	float4 Color=c_baseColor;
	float4 normal=d_GetNormal(pos);

	if (c_Illustrative)
	{
		float n_dot_c;
		n_dot_c=dot(normal,CameraDirection);
		float rate=min(powf(1.0f-n_dot_c*n_dot_c,c_specular_power*0.5f),1.0f);
		Color.x=Color.x*rate+c_BackgroundColor.x*(1.0f-rate);
		Color.y=Color.y*rate+c_BackgroundColor.y*(1.0f-rate);
		Color.z=Color.z*rate+c_BackgroundColor.z*(1.0f-rate);
	}
	else
	{
		float n_dot_l;
		float n_dot_h; 

		float3 k=c_ambient;
		float3 b=make_float3(0.0f); 
		n_dot_l=dot(normal,c_light_direction);
		
		if ((n_dot_l>0.0f && increasing) || (n_dot_l<0.0f && !increasing))
		{
			n_dot_l=n_dot_l>0?n_dot_l:-n_dot_l;
			k+=c_diffuse*n_dot_l;
			float4 HalfwayVector=normalize(c_light_direction+CameraDirection); //38
			n_dot_h=dot(normal,HalfwayVector);
			n_dot_h=n_dot_h>0?n_dot_h:-n_dot_h;
			if (n_dot_h > c_spec_threshold)
			{
				b=c_specular*pow(n_dot_h,c_specular_power);
			}
		}

		Color.x=Color.x*k.x+b.x;
		Color.y=Color.y*k.y+b.y;
		Color.z=Color.z*k.z+b.z;
	}
	return Color;
}


static __device__
float d_func(float x,float a,float b,float c,float d)
{
	return x*(x*(x*a+b)+c)+d;
}
static __device__
float d_secant_search(float x1,float x2, float y1,float y2,float a,float b,float c,float d)
{
	float err=(x2-x1)/10000.0f;
	bool increasing=(y1<=y2);

	float x_zero=(0.0f-y1)/(y2-y1)*(x2-x1)+x1;
	float x_zero1;
	
	do
	{
		float y=d_func(x_zero,a,b,c,d);

		if ((increasing && y>=0) || (!increasing && y<=0))
		{
			x2=x_zero;
			y2=y;
		}
		else
		{
			x1=x_zero;
			y1=y;
		}
		x_zero1=x_zero;
		x_zero=(0.0f-y1)/(y2-y1)*(x2-x1)+x1;
	}
	while(fabsf(x_zero1-x_zero)>err);
	return x_zero;

}

static __device__
bool d_Solve_Cubic_Eq(float a,float b,float c,float d,float& root)
{
	float x1=0.0f;
	float x2=1.0f;
	float y1=d;
	float y2=a+b+c+d;

	if (a==0.0f)
	{
		if (b!=0.0f)
		{
			float x=-c/(2.0f*b);
			float y=d_func(x,a,b,c,d);
			if (x>x1 && x<x2) 
			{
				if (signbit(y)==signbit(y1))
				{
					x1=x;
					y1=y;
				}
				else
				{
					x2=x;
					y2=y;
				}
			}
		}
	}
	else
	{
		float _b=b/(3.0f*a);
		float _c=c/(3.0f*a);
		
		float delta=_b*_b-_c;		
		if (delta>0)
		{
			delta=sqrtf(delta);
			float r1,r2;
			if (_b>0.0f) 
			{
				r1=-_b-delta;
				r2=_c/r1;
			}
			else 
			{
				r2=-_b+delta;
				r1=_c/r2;
			}

			float fr1=d_func(r1,a,b,c,d);
			float fr2=d_func(r2,a,b,c,d);
			
			if (r1>x1 && r1<x2) 
			{
				if (signbit(fr1)==signbit(y1))
				{
					x1=r1;
					y1=fr1;
				}
				else
				{
					x2=r1;
					y2=fr1;
				}
			}
			if (r2>x1 && r2<x2) 
			{
				if (signbit(fr2)==signbit(y1))
				{
					x1=r2;
					y1=fr2;
				}
				else
				{
					x2=r2;
					y2=fr2;
				}
			}
		}
	}
	
	if (signbit(y1)==signbit(y2)) return false;

	root=d_secant_search(x1,x2,y1,y2,a,b,c,d);

	return true;

}

static bool IsosurfaceRayCasterPrepare(BasicLightParam* lighting, IsosurfaceRayCasterParam *RayCaster)
{
	if (!RayCaster) return false;

	float isovalue=RayCaster->m_Isovalue;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_isovalue, &isovalue, sizeof(float),0,cudaMemcpyHostToDevice));

	float baseColor[4];
	baseColor[0]=RayCaster->m_SurfaceColor[0];
	baseColor[1]=RayCaster->m_SurfaceColor[1];
	baseColor[2]=RayCaster->m_SurfaceColor[2];
	baseColor[3]=1.0f;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_baseColor, baseColor, sizeof(float)*4,0,cudaMemcpyHostToDevice));
	
	float lightColor[3];
	lightColor[0]=lighting->m_LightColor[0];
	lightColor[1]=lighting->m_LightColor[1];
	lightColor[2]=lighting->m_LightColor[2];
	
	float LightIntensity=lighting->m_LightIntensity;

	lightColor[0]*=LightIntensity;
	lightColor[1]*=LightIntensity;
	lightColor[2]*=LightIntensity;

	float f_ambient=lighting->m_Ambient;
	float ambient[3];
	ambient[0]=f_ambient*lightColor[0];
	ambient[1]=f_ambient*lightColor[1];
	ambient[2]=f_ambient*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_ambient, ambient, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float f_diffuse=lighting->m_Diffuse;
	float diffuse[3];
	diffuse[0]=f_diffuse*lightColor[0];
	diffuse[1]=f_diffuse*lightColor[1];
	diffuse[2]=f_diffuse*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_diffuse, diffuse, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float f_specular=lighting->m_Specular;
	float specular[3];
	specular[0]=f_specular*lightColor[0];
	specular[1]=f_specular*lightColor[1];
	specular[2]=f_specular*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_specular, specular, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float specular_power=lighting->m_SpecularPower;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_specular_power, &specular_power, sizeof(float),0,cudaMemcpyHostToDevice));

	float spec_threshold=powf(0.001f/max(max(specular[0],specular[1]),specular[2]),1.0f/specular_power);
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_spec_threshold, &spec_threshold, sizeof(float),0,cudaMemcpyHostToDevice));

	float light_direction[4];
	light_direction[0]=lighting->m_LightDirection[0];
	light_direction[1]=lighting->m_LightDirection[1];
	light_direction[2]=lighting->m_LightDirection[2];
	light_direction[3]=0.0f;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_light_direction, light_direction, sizeof(float)*4,0,cudaMemcpyHostToDevice));

	int illustrative=lighting->m_Illustrative?1:0;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_Illustrative, &illustrative, sizeof(int),0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_BackgroundColor, lighting->m_BackgroundColor, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	return true;
}