#include "BasicLightParam.h"
#include "IsosurfaceRayCasterParam.h"
#include "Vector4.h"
#include "Vector3.h"


static inline float GetScalarValue(int x,int y,int z);
static inline float GetScalarValue(Vector4 pos);
static inline Vector4 GetNormal(const Vector4& pos);

static float isovalue;
static Vector4 baseColor;
static Vector3 ambient;
static Vector3 diffuse;
static Vector3 specular;
static float specular_power;
static Vector4 light_direction;
static float spec_threshold;

static bool illustrative;
static Vector3 backgroundColor;

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif


static inline Vector4 GetColor(const Vector4& pos,const Vector4& CameraDirection,bool increasing)
{
	Vector4 Color=baseColor;
	Vector4 normal=GetNormal(pos);

	if (illustrative)
	{
		float n_dot_c;
		n_dot_c=normal*CameraDirection;
		float rate=min(powf(1.0f-n_dot_c*n_dot_c,specular_power*0.5f),1.0f);
		Color[0]=Color[0]*rate+backgroundColor[0]*(1.0f-rate);
		Color[1]=Color[1]*rate+backgroundColor[1]*(1.0f-rate);
		Color[2]=Color[2]*rate+backgroundColor[2]*(1.0f-rate);
	}
	else
	{

		float n_dot_l;
		float n_dot_h;

		Vector3 k=ambient;
		Vector3 b;
		n_dot_l=normal*light_direction;
		if ((n_dot_l>0.0f && increasing) || (n_dot_l<0.0f && !increasing))
		{
			n_dot_l=n_dot_l>0?n_dot_l:-n_dot_l;
			k+=diffuse*n_dot_l;
			Vector4 HalfwayVector=light_direction+CameraDirection;
			HalfwayVector.Normalize();
			n_dot_h=normal*HalfwayVector;
			if (n_dot_h<0) n_dot_h=-n_dot_h;
			if (n_dot_h > spec_threshold)
			{
				b=specular*powf(n_dot_h,specular_power);
			}
		}
		Color[0]=k[0]*Color[0]+b[0];
		Color[1]=k[1]*Color[1]+b[1];
		Color[2]=k[2]*Color[2]+b[2];
	}
	return Color;
}
static inline float func(float x,float a,float b,float c,float d)
{
	return x*(x*(x*a+b)+c)+d;
}
static inline float secant_search(float x1,float x2, float y1,float y2,float a,float b,float c,float d)
{
	float err=(x2-x1)/10000.0f;
	bool increasing=(y1<=y2);

	float x_zero=(0.0f-y1)/(y2-y1)*(x2-x1)+x1;
	float x_zero1;
	
	do
	{
		float y=func(x_zero,a,b,c,d);

		if ((increasing && y>=0) || (!increasing && y<=0))
		{
			x2=x_zero;
			y2=y;
		}
		else
		{
			x1=x_zero;
			y1=y;
		}
		x_zero1=x_zero;
		x_zero=(0.0f-y1)/(y2-y1)*(x2-x1)+x1;
	}
	while(fabsf(x_zero1-x_zero)>err);
	return x_zero;
}

#ifndef signbit
static inline int signbit(float x)
{
	if (x>0) return 1;
	else if (x<0) return -1;
	else return 0;
}
#endif

static inline bool Solve_Cubic_Eq(float a,float b,float c,float d,float x1,float x2,float& root)
{
	float y1=func(x1,a,b,c,d);
	float y2=func(x2,a,b,c,d);

	if (a==0.0f)
	{
		if (b!=0.0f)
		{
			float x=-c/(2.0f*b);
			float y=func(x,a,b,c,d);
			if (x>x1 && x<x2) 
			{
				if (signbit(y)==signbit(y1))
				{
					x1=x;
					y1=y;
				}
				else
				{
					x2=x;
					y2=y;
				}
			}
		}
	}
	else
	{
		float _b=b/(3.0f*a);
		float _c=c/(3.0f*a);
		
		float delta=_b*_b-_c;		
		if (delta>0)
		{
			delta=sqrtf(delta);
			float r1,r2;
			if (_b>0.0f) 
			{
				r1=-_b-delta;
				r2=_c/r1;
			}
			else 
			{
				r2=-_b+delta;
				r1=_c/r2;
			}

			float fr1=func(r1,a,b,c,d);
			float fr2=func(r2,a,b,c,d);
			
			if (r1>x1 && r1<x2) 
			{
				if (signbit(fr1)==signbit(y1))
				{
					x1=r1;
					y1=fr1;
				}
				else
				{
					x2=r1;
					y2=fr1;
				}
			}
			if (r2>x1 && r2<x2) 
			{
				if (signbit(fr2)==signbit(y1))
				{
					x1=r2;
					y1=fr2;
				}
				else
				{
					x2=r2;
					y2=fr2;
				}
			}
		}
	}
	
	if (signbit(y1)==signbit(y2)) return false;

	root=secant_search(x1,x2,y1,y2,a,b,c,d);

	return true;
}


static bool IsosurfaceRayCasterPrepare(BasicLightParam* lighting, IsosurfaceRayCasterParam *RayCaster)
{
	if (!RayCaster) return false;

	isovalue=RayCaster->m_Isovalue;
	baseColor[0]=RayCaster->m_SurfaceColor[0];
	baseColor[1]=RayCaster->m_SurfaceColor[1];
	baseColor[2]=RayCaster->m_SurfaceColor[2];
	baseColor[3]=1.0f;
	
	float lightColor[3];
	lightColor[0]=lighting->m_LightColor[0];
	lightColor[1]=lighting->m_LightColor[1];
	lightColor[2]=lighting->m_LightColor[2];

	float LightIntensity=lighting->m_LightIntensity;

	lightColor[0]*=LightIntensity;
	lightColor[1]*=LightIntensity;
	lightColor[2]*=LightIntensity;

	float f_ambient=lighting->m_Ambient;

	ambient[0]=f_ambient*lightColor[0];
	ambient[1]=f_ambient*lightColor[1];
	ambient[2]=f_ambient*lightColor[2];

	float f_diffuse=lighting->m_Diffuse;

	diffuse[0]=f_diffuse*lightColor[0];
	diffuse[1]=f_diffuse*lightColor[1];
	diffuse[2]=f_diffuse*lightColor[2];

	float f_specular=lighting->m_Specular;

	specular[0]=f_specular*lightColor[0];
	specular[1]=f_specular*lightColor[1];
	specular[2]=f_specular*lightColor[2];

	specular_power=lighting->m_SpecularPower;

	spec_threshold=powf(0.001f/max(max(specular[0],specular[1]),specular[2]),1.0f/specular_power);

	light_direction[0]=lighting->m_LightDirection[0];
	light_direction[1]=lighting->m_LightDirection[1];
	light_direction[2]=lighting->m_LightDirection[2];
	light_direction[3]=0.0f;

	::illustrative=lighting->m_Illustrative;
	::backgroundColor[0]=lighting->m_BackgroundColor[0];
	::backgroundColor[1]=lighting->m_BackgroundColor[1];
	::backgroundColor[2]=lighting->m_BackgroundColor[2];

	return true;
}


